;; -*- mode: scheme; coding: utf-8 -*-
;; Copyright (c) 2024 Vadym Kochan
;; SPDX-License-Identifier: MIT
#!r6rs

(library (image png)
  (export 
          PNG-COLOR-TYPE-GRAY       
          PNG-COLOR-TYPE-RGB        
          PNG-COLOR-TYPE-INDEXED    
          PNG-COLOR-TYPE-GRAY-ALPHA 
          PNG-COLOR-TYPE-RGBA

          read-png-chunks
          write-png-chunks
          make-png-chunk
          make-ihdr-chunk
          make-idat-chunk
          png-chunk-type
          png-chunk-data
          png-chunk-length

          make-png
          read-png
          write-png
          png?
          png-height
          png-width
          png-color-bits
          png-color-type
          png-chunks
          png-pixel-size
          png-color-channels
          png-data)
  (import (rnrs)
          (hashing crc)
          (image private zlib))

(define PNG-COLOR-TYPE-GRAY       0)
(define PNG-COLOR-TYPE-RGB        2)
(define PNG-COLOR-TYPE-INDEXED    3)
(define PNG-COLOR-TYPE-GRAY-ALPHA 4)
(define PNG-COLOR-TYPE-RGBA       6)

(define ascii-encoder (make-transcoder (latin-1-codec)))

(define-crc crc-32)

(define string->bytes
  (lambda (str)
    (string->bytevector str ascii-encoder)))

(define-record-type png-chunk
  (fields length type data)
  (protocol
    (lambda (new)
      (lambda (type data)
        (new (bytevector-length data) type data)))))

(define (symbol->color-channels sym)
  (case sym
    ((gray) 1)
    ((rgb) 3)
    ((indexed) 1)
    ((gray-alpha) 2)
    ((rgba) 4)
    (else 0)))

(define-record-type png
  (fields width
          height
          color-bits
          color-type
          chunks
          data
          pixel-size
          color-channels)
  (protocol
    (lambda (new)
      (case-lambda
        [(width height bits type data)
         (let ([color-channels (symbol->color-channels type)])
           (new width height bits type
                #f data
                (fx* color-channels (fxdiv bits 8))
                color-channels))]

        [(width height bits type data chunks)
         (let ([color-channels (symbol->color-channels type)])
           (new width height bits type chunks data
                (fx* color-channels (fxdiv bits 8))
                color-channels))] ))))

(define PNG-SIGNATURE #vu8(137 80 78 71 13 10 26 10))

(define iend (make-png-chunk "IEND" (make-bytevector 0)))

(define (check-signature port)
  (let ([bytes (get-bytevector-n port (bytevector-length PNG-SIGNATURE))])
    (when (not (equal? bytes PNG-SIGNATURE))
      (error 'read-png "Bad PNG signature"))))

(define (make-word-reader port buf)
  (let ([bytes buf] [port port])
    (lambda ()
      (let ([eof? (get-bytevector-n! port bytes 0 4)])
        (if (eof-object? eof?)
            (error 'read-word "Unexpected EOF")
            (bytevector-u32-ref bytes 0 (endianness big)))))))

(define (make-word-writer port)
  (let ([bytes (make-bytevector 4 0)] [port port])
    (lambda (val)
      (bytevector-u32-set! bytes 0 val (endianness big))
      (put-bytevector port bytes))))

(define (chunk-calc-crc chk)
  (let* ([crc (crc-32-update (crc-32-init) (string->bytes (png-chunk-type chk)))]
         [crc (crc-32-update crc (png-chunk-data chk))]
         [crc (crc-32-finish crc)])
    crc))

(define (check-chunk-crc type orig-crc bv)
  (let* ([crc (crc-32-update (crc-32-init) (string->bytes type))]
         [crc (crc-32-update crc bv)]
         [calc-crc (crc-32-finish crc)])
    (when (not (equal? calc-crc orig-crc))
      (error 'check-crc32 (string-append
                            "Chunk " type " has crc32 mismatch: " (number->string calc-crc 16)
                            " (orig: " (number->string orig-crc 16) ")")))))

(define read-png-chunks
  (lambda (port)
    (check-signature port)
    (let* ([bytes (make-bytevector 4)]
           [read-word (make-word-reader port bytes)])
      (let loop ([chunks (list)] [end? #f])
        (if end?
            chunks
            (let* ([size (read-word)]
                   [type (let () (read-word) (utf8->string bytes))]
                   [data (get-bytevector-n port size)]
                   [crc (read-word)]
                   [chk (make-png-chunk type data)])
              (check-chunk-crc type crc data)
              (loop (append chunks (list chk))
                    (equal? type "IEND"))))))))

(define (write-png-chunks chunks port)
  (let ([ihdr #f]
        [plte #f]
        [idats (list)]
        [end #f])
    (filter
      (lambda (chk)
        (cond
          ((equal? (png-chunk-type chk) "IHDR")
           (set! ihdr chk)
           #f)
          ((equal? (png-chunk-type chk) "PLTE")
           (set! plte chk)
           #f)
          ((equal? (png-chunk-type chk) "IDAT")
           (set! idats (append idats (list chk)))
           #f)
          ((equal? (png-chunk-type chk) "IEND")
           (set! end chk)
           #f)
          (else #t)))
      chunks)
    (put-bytevector port PNG-SIGNATURE)
    (when ihdr
      (write-chunk ihdr port))
    (when plte
      (write-chunk plte port))
    (for-each (lambda (idat) (write-chunk idat port))
      idats)
    (write-chunk (or end iend) port)))

(define (chunk-u8-ref chk offs)
  (bytevector-u8-ref (png-chunk-data chk) offs))

(define (chunk-u32-ref chk offs)
  (bytevector-u32-ref (png-chunk-data chk) offs (endianness big)))

(define (chunk-u8-set! chk offs val)
  (bytevector-u8-set! (png-chunk-data chk) offs val))

(define (chunk-u32-set! chk offs val)
  (bytevector-u32-set! (png-chunk-data chk) offs val (endianness big)))

(define (png-color-type->symbol color-type)
  (case color-type
    ((0) 'gray)
    ((2) 'rgb)
    ((3) 'indexed)
    ((4) 'gray-alpha)
    ((6) 'rgba)
    (else 'unknown)))

(define (symbol->color-type sym)
  (case sym
    ((gray) 0)
    ((rgb) 2)
    ((indexed) 3)
    ((gray-alpha) 4)
    ((rgba) 6)
    (else (error 'symbol->color-type "Invalid color type symbol: " sym))))

(define (get-chunk type chunks)
  (find (lambda (chk)
          (equal? (png-chunk-type chk) type))
    chunks))

(define (calc-pixel-size bits type)
  (let ([color-channels (symbol->color-channels (png-color-type->symbol type))])
    (fx* color-channels (fxdiv bits 8))))

(define (paeth-predictor a b c)
  (let* ([p (fx- (fx+ a b) c)]
         [pa (abs (fx- p a))]
         [pb (abs (fx- p b))]
         [pc (abs (fx- p c))])
    (cond
      [(and (fx<=? pa pb) (fx<=? pa pc)) a]
      [(fx<=? pb pc) b]
      [else c])))

(define (unfilter-pixels width height px-size data)
  (define scanline-size (fx* width px-size))
  (define unfilt-data (make-bytevector (* width height px-size)))

  (define (get-filter-values p y)
    (let ([x (bytevector-u8-ref data (+ (fx* y (fx+ 1 scanline-size)) 1 p))]
          [a (if (fx>=? p px-size)
                 (bytevector-u8-ref unfilt-data (fx- (fx+ (fx* scanline-size y) p) px-size)) 
                 0)]
          [b (if (fx>? y 0)
                 (bytevector-u8-ref unfilt-data (fx+ (fx* scanline-size (fx- y 1)) p))
                 0)]
          [c (if (and (fx>? y 0) (fx>=? p px-size))
                 (bytevector-u8-ref unfilt-data (fx- (fx+ (fx* scanline-size (fx- y 1)) p) px-size))
                 0)])
      (values x a b c)))

  (let loop-y ([y 0])
    (when (fx<? y height)
      (let ([filter-id (bytevector-u8-ref data (fx* (fx+ scanline-size 1) y))])
        (let loop-p ([p 0])
          (when (fx<? p scanline-size)
            ;; x - current byte
            ;; a - left pixel byte (relative to current) 
            ;; b - above pixel byte (relative to current) 
            ;; c - left pixel byte (relative to b)
            (let-values ([(x a b c) (get-filter-values p y)])
              (let* ([out (case filter-id
                            ((0) x) ;; Filter None (usually used for indexed images)
                            ((1) (fx+ x a)) ;; Filter Sub
                            ((2) (fx+ x b)) ;; Filter Up
                            ((3) (fx+ x (fxarithmetic-shift-right (fx+ a b) 1))) ;; Filter Average
                            ((4) (paeth-predictor a b c))
                            (else 0))]
                     [out (bitwise-and out #xFF)])
                (bytevector-u8-set! unfilt-data (fx+ (fx* scanline-size y) p) out)))
            (loop-p (fx+ p 1))))
        (loop-y (fx+ y 1)))))
  unfilt-data)

(define (collect-idat-chunks chunks)
  (let-values ([(outp get-bytes) (open-bytevector-output-port)])
    (call-with-port outp
      (lambda (outp)
        (for-each (lambda (chk)
                    (when (equal? (png-chunk-type chk) "IDAT")
                      (put-bytevector outp (png-chunk-data chk))))
          chunks)
        (get-bytes)))))

(define (read-png port)
  (let* ([chunks (read-png-chunks port)]
         [ihdr (list-ref chunks 0)]
         [idat (get-chunk "IDAT" chunks)]
         [width (chunk-u32-ref ihdr 0)]
         [height (chunk-u32-ref ihdr 4)]
         [color-bits (chunk-u8-ref ihdr 8)]
         [color-type (chunk-u8-ref ihdr 9)]
         [data (if idat (collect-idat-chunks chunks) #f)]
         [data (if idat (zlib-decompress data) #f)]
         [data (if data (unfilter-pixels width height (calc-pixel-size color-bits color-type) data) #f)])
    (make-png width height
                    color-bits
                    (png-color-type->symbol color-type)
                    data chunks)))

(define (write-chunk chk port)
  (let ([write-word (make-word-writer port)])
    (write-word (png-chunk-length chk))
    (put-bytevector port (string->bytes (png-chunk-type chk)))
    (put-bytevector port (png-chunk-data chk))
    (write-word (chunk-calc-crc chk))))

(define (make-ihdr-chunk width height bits color-type)
  (let ([ihdr (make-png-chunk "IHDR" (make-bytevector 13 0))])
    (chunk-u32-set! ihdr 0 width)
    (chunk-u32-set! ihdr 4 height)
    (chunk-u8-set! ihdr 8 bits)
    (chunk-u8-set! ihdr 9 color-type)
    ihdr))

(define (filter-pixels width height px-size data)
  (let* ([scanline-size (fx* width px-size)]
         [filt-data (make-bytevector (fx* (fx+ scanline-size 1) height))])
    (let loop-y ([y 0])
      (when (fx<? y height)
        ;; no filter yet
        (bytevector-u8-set! filt-data (fx* (fx+ scanline-size 1) y) 0)
        (let loop-c ([c 0])
          (when (fx<? c scanline-size)
            (let ([val (bytevector-u8-ref data (fx+ (fx* scanline-size y) c))])
              (bytevector-u8-set! filt-data (+ (fx* (fx+ scanline-size 1) y) c 1) val)
              (loop-c (fx+ c 1)))))
        (loop-y (fx+ y 1))))
    filt-data))

(define (make-idat-chunk width height px-size data)
  (let* ([filt (filter-pixels width height px-size data)]
         [comp (zlib-compress filt)]
         [idat (make-png-chunk "IDAT" comp)])
    idat))

(define (write-png img port)
  (let* ([width (png-width img)]
         [height (png-height img)]
         [bits (png-color-bits img)]
         [color-type (symbol->color-type (png-color-type img))]
         [px-size (png-pixel-size img)]
         [ihdr (make-ihdr-chunk width height bits color-type)]
         [idat (make-idat-chunk width height px-size (png-data img))])
    (put-bytevector port PNG-SIGNATURE)
    (write-chunk ihdr port)
    (write-chunk idat port)
    (write-chunk iend port)))

)
