;; -*- mode: scheme; coding: utf-8 -*-
;; Copyright (c) 2024 Vadym Kochan
;; SPDX-License-Identifier: MIT
#!r6rs

(library (image private zlib)
  (export
    zlib-compress
    zlib-decompress)
  (import
    (rnrs)
    (compression zlib)
    (compression adler-32))

(define ZLIB-MIN-HEADER-SIZE  2)
(define ZLIB-FOOTER-SIZE      4)
(define ZLIB-MIN-OVERHEAD     (fx+ ZLIB-MIN-HEADER-SIZE ZLIB-FOOTER-SIZE))
(define ZLIB-CM-DEFLATE       8)
(define ZLIB-CINFO-32K-WINDOW 7)

(define ZLIB-FASTEST-COMPRESSION 0)
(define ZLIB-FAST-COMPRESSION     1)
(define ZLIB-DEFAULT-COMPRESSION 2)
(define ZLIB-SLOWEST-COMPRESSION 4)

(define ZLIB-BLOCKTYPE-UNCOMPRESSED    0)
(define ZLIB-BLOCKTYPE-STATIC-HUFFMAN  1)
(define ZLIB-BLOCKTYPE-DYNAMIC-HUFFMAN 2)

(define (put-le16 outp val)
  (put-u8 outp (bitwise-bit-field val 0 8))
  (put-u8 outp (bitwise-bit-field val 8 16)))

(define (put-be16 outp val)
  (put-u8 outp (bitwise-bit-field val 8 16))
  (put-u8 outp (bitwise-bit-field val 0 8)))

(define (put-be32 outp val)
  (put-u8 outp (bitwise-bit-field val 24 32))
  (put-u8 outp (bitwise-bit-field val 16 24))
  (put-u8 outp (bitwise-bit-field val 8 16))
  (put-u8 outp (bitwise-bit-field val 0 8)))

(define blk-size 32768)

(define (deflate-compress-none inp outp)
  (define buf (make-bytevector blk-size 0))

  (let loop ([len (get-bytevector-n! inp buf 0 blk-size)])
    (when (not (eof-object? len))
      (let* ([last? (fx<? len blk-size)]
             [bfinal (if last? 1 0)]
             [btype (fxarithmetic-shift ZLIB-BLOCKTYPE-UNCOMPRESSED 1)])
        (put-u8 outp (bitwise-ior bfinal btype))
        (put-le16 outp len)
        (put-le16 outp (fxnot len))
        (put-bytevector outp buf 0 len)
        (when (not last?)
          (loop (get-bytevector-n! inp buf 0 blk-size)))))))

(define (zlib-compress bv)
  (let* ([hdr (bitwise-ior (fxarithmetic-shift ZLIB-CM-DEFLATE 8)
                           (fxarithmetic-shift ZLIB-CINFO-32K-WINDOW 12))]
         [lvl-comp 0]
         [lvl-hint (cond ((< lvl-comp 2) ZLIB-FASTEST-COMPRESSION)
                         ((< lvl-comp 6) ZLIB-FAST-COMPRESSION)
                         ((< lvl-comp 8) ZLIB-DEFAULT-COMPRESSION)
                         (else ZLIB-SLOWEST-COMPRESSION))]
         [hdr (bitwise-ior hdr (fxarithmetic-shift lvl-hint 6))]
         [hdr (bitwise-ior hdr (fx- 31 (fxmod hdr 31)))]
         [inp (open-bytevector-input-port bv)])
    (let-values ([(outp get-bytes) (open-bytevector-output-port)])
      (put-be16 outp hdr)
      (deflate-compress-none inp outp)
      (put-be32 outp (adler-32 bv))
      (let ([bytes (get-bytes)])
        (close-port outp)
        (close-port inp)
        bytes))))

(define (zlib-decompress bv)
  (call-with-port
    (make-zlib-input-port (open-bytevector-input-port bv) "png" #f 'close-it '())
    (lambda (data)
      (get-bytevector-all data))))

)
